import shlex
import subprocess

from agent.exceptions.agent_error import AgentError


def execute_shell_command(command):
    try:
        subprocess.call(shlex.split(command))
    except subprocess.CalledProcessError as err:
        raise AgentError(expression=command, message=err.output)



import json
import os
from pathlib import Path
from typing import Dict

from agent.commands import logger
from agent.config.properties import BASE_SERVICES_PATH
from agent.constants import DOCKER_COMPOSE_FILE
from agent.exceptions.agent_error import AgentError


def get_base_service_path(service_id):
    return f'{BASE_SERVICES_PATH}/{service_id}'


def get_docker_compose_path(service_id):
    return f'{get_base_service_path(service_id)}/{DOCKER_COMPOSE_FILE}'


def get_field_from_body(key, body):
    try:
        return body[key]
    except KeyError as e:
        logger.error(f'{body} does not contain any {key}!')
        raise AgentError(body, e)


def prepare_message_to_send(message_type, body):
    message = dict(messageType=message_type, body=body)
    return json.dumps(message).encode('utf-8')


def save_config(config: Dict[str, str], service_id: str) -> None:
    config_directory = _create_config_path_for_service(service_id)
    file_path = f'{config_directory}/config.json'
    _save_data_to_file(config, file_path)
    logger.info(f'Config updated: {file_path}')


def _save_data_to_file(data, file_path):
    with open(file_path, 'w') as f:
        json.dump(data, f)


def _create_config_path_for_service(service_id):
    service_base_path = get_base_service_path(service_id)
    config_path = f'{service_base_path}/config'
    _create_config_path_if_needed(config_path)
    return config_path


def _create_config_path_if_needed(config_path):
    if not os.path.isdir(config_path):
        logger.info(f'Directory {config_path} does not exist, creating one')
        try:
            Path(config_path).mkdir(parents=True, exist_ok=True)
        except PermissionError as e:
            logger.error(f'User has no permission to this path: {config_path}')
            raise AgentError(config_path, e)



from agent.commands import logger
from agent.commands.shell_executor import execute_shell_command
from agent.exceptions.agent_error import AgentError


def docker_compose_pull(path):
    execute_shell_command(f'docker-compose -f {path} pull')


def docker_compose_up(path):
    execute_shell_command(f'docker-compose -f {path} up -d')


def docker_compose_down(path):
    execute_shell_command(f'docker-compose -f {path} down')


def docker_compose_restart(path):
    try:
        execute_shell_command(f'docker-compose -f {path} restart')
    except AgentError:
        logger.exception(f'Failed to restart docker-compose {path}')
        docker_compose_up(path)



from autobahn.twisted.websocket import WebSocketClientProtocol, WebSocketClientFactory
from twisted.internet import reactor
from twisted.internet.protocol import ReconnectingClientFactory

from agent.rabbitmq.client import ReconnectingRabbitMqClient
from agent.websocket import logger
from agent.websocket.utils import resolve_ws_endpoint


class AgentWebSocketClient(WebSocketClientProtocol):
    _instance = None

    @staticmethod
    def init(*args):  # AgentWebSocketFactory requires method with arguments
        AgentWebSocketClient._instance = AgentWebSocketClient()
        return AgentWebSocketClient._instance

    @staticmethod
    def get_instance():
        if not AgentWebSocketClient._instance:
            return AgentWebSocketClient.init()
        return AgentWebSocketClient._instance

    def __init__(self):
        super().__init__()
        self.on_message_handler = None

    def set_on_message_handler(self, on_message_handler):
        self.on_message_handler = on_message_handler

    def onConnect(self, response):
        from agent.communication.message_processor import process_message
        from agent.scheduler.scheduled_tasks_manager import ScheduledTaskManager

        logger.info(f"Connected to Server: {response.peer}")
        self.set_on_message_handler(process_message)
        scheduled_task_manager = ScheduledTaskManager.get_instance()
        scheduled_task_manager.start_executing()
        ReconnectingRabbitMqClient.get_instance().start_consuming_messages()

    def connectionLost(self, reason):
        from agent.scheduler.scheduled_tasks_manager import ScheduledTaskManager

        ScheduledTaskManager.get_instance().stop_executing()
        ReconnectingRabbitMqClient.get_instance().stop_consuming_messages()
        super().connectionLost(reason)

    def onMessage(self, payload, isBinary):
        logger.info(f"Received a message: {payload}")
        if self.on_message_handler:
            reactor.callInThread(self.on_message_handler, payload)

    def safe_message_send(self, message):
        try:
            self.sendMessage(message)
            logger.info(f"Message sent: {message}")
        except Exception as e:
            logger.error(f"Client connection lost because of: {e}.. retrying ..")
            AgentWebSocketFactory.get_instance().retry()


class AgentWebSocketFactory(WebSocketClientFactory, ReconnectingClientFactory):
    protocol = AgentWebSocketClient.init
    _instance = None
    maxDelay = 10.0

    @staticmethod
    def get_instance():
        if not AgentWebSocketFactory._instance:
            AgentWebSocketFactory._instance = AgentWebSocketFactory(
                resolve_ws_endpoint()
            )
        return AgentWebSocketFactory._instance

    def clientConnectionFailed(self, connector, reason):
        logger.error(f"Client connection failed because of: {reason}.. retrying ..")
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        logger.error(f"Client connection lost because of: {reason}.. retrying ..")
        self.retry(connector)



import requests
from requests import HTTPError

from agent.certificates import logger
from agent.exceptions.agent_error import AgentError


class CertificateClient:
    def __init__(self, base_path, access_token):
        self.base_path = base_path
        self.access_token = access_token

    def send_csr(self, service_id, csr, cert_type):
        logger.info('Sending csr')
        csr_files = {'csr': ('service.csr', csr)}
        url = f'{self.base_path}/v1/certificates/{service_id}?type={cert_type}'

        try:
            resp = requests.post(url, files=csr_files, headers={'Authorization': f'Bearer {self.access_token}'})
            resp.raise_for_status()
        except HTTPError as error:
            logger.error(f'Error during certificate generation: {error}')
            raise AgentError(expression=url, message=error)
        return resp.text

    def get_ca_certs_chain(self):
        resp = requests.get(f'{self.base_path}/v1/certificates',
                            headers={'Authorization': f'Bearer {self.access_token}'})
        return resp.text



from autobahn.twisted.websocket import connectWS
from twisted.internet import reactor

from agent.logger import setup_logger
from agent.rabbitmq.setup import setup_rabbitmq
from agent.websocket.client import AgentWebSocketFactory

logger = setup_logger()


def main():
    logger.info("Launching Agent...")
    setup_rabbitmq()
    connectWS(AgentWebSocketFactory.get_instance())
    reactor.run()


if __name__ == "__main__":
    main()



import json
from collections import namedtuple

from agent.commands import dispatcher
from agent.communication import logger
from agent.exceptions.agent_error import AgentError
from agent.rabbitmq.publisher import publish_message

command = namedtuple("command", ["type", "body"])


def process_message(payload: bytes) -> None:
    data = convert_message_to_json(payload.decode("utf8"))
    if data.get("nodeServiceId"):
        publish_message(data)
    else:
        process_json_message(data)


def process_json_message(json_message):
    try:
        resolved_command = process_message_to_command(json_message)
        dispatcher.dispatch_command(resolved_command.type, resolved_command.body)
    except AgentError as error:
        logger.error(f"Dispatching command failed due to the: {error.message}")


def process_message_to_command(json_message):
    message_type = json_message["messageType"]
    body = json_message["body"]
    return command(type=message_type, body=body)


def convert_message_to_json(message):
    try:
        json_message = json.loads(message)
    except json.decoder.JSONDecodeError as e:
        logger.error(f"{message} is not json!")
        raise AgentError(message, e)
    return json_message



from pathlib import Path

from OpenSSL import crypto
from OpenSSL.crypto import PKey, TYPE_RSA, X509Req, FILETYPE_PEM, PKCS12

from agent.certificates import logger

CERT_BEGIN_PREFIX = '-----BEGIN CERTIFICATE-----'
CERT_END_PREFIX = '-----END CERTIFICATE-----'

KNOWN_PASSWORD = 'known-password'


def parse_ca_certs_chain(ca_certs_chain):
    ca_certs = ca_certs_chain.split(f'{CERT_END_PREFIX}\n{CERT_BEGIN_PREFIX}')
    return list(map(convert_to_certificate, ca_certs))


def convert_to_certificate(ca_cert):
    if not ca_cert.startswith(CERT_BEGIN_PREFIX):
        ca_cert = CERT_BEGIN_PREFIX + ca_cert
    if not ca_cert.endswith(CERT_END_PREFIX):
        ca_cert = ca_cert + CERT_END_PREFIX

    return crypto.load_certificate(FILETYPE_PEM, ca_cert)


class CertificateConfigGenerator:
    def __init__(self, service_configs_path, cert_client):
        self.service_configs_path = service_configs_path
        self.cert_client = cert_client
        self.ca_certs_chain = self.cert_client.get_ca_certs_chain()
        self.ca_certs = parse_ca_certs_chain(self.ca_certs_chain)
        self.key = PKey()
        self.key.generate_key(TYPE_RSA, 2048)

    def generate_cert_config(self, service_id):
        self._generate_cert_config_for_type(service_id, 'SERVER')
        self._generate_cert_config_for_type(service_id, 'CLIENT')

    def _generate_cert_config_for_type(self, service_id, cert_type):
        self.cert_type = cert_type
        Path(self._service_config_path(service_id)).mkdir(parents=True, exist_ok=True)
        self._save_private_key(service_id)
        self._save_ca_certs_chain(service_id)
        cert = self._generate_cert(service_id)
        self._generate_p12_cert(cert, service_id)

    def _save_ca_certs_chain(self, service_id):
        certs_chain_path = self._ca_certs_chain_path(service_id)
        logger.info(f'Saving certificates chain: {certs_chain_path}')
        with open(certs_chain_path, 'w') as cert_file:
            cert_file.write(self.ca_certs_chain)

        self._generate_truststore(service_id)

    def _save_private_key(self, service_id):
        private_key_path = self._private_key_path(service_id)
        logger.info(f'Saving private key: {private_key_path}')

        with open(private_key_path, 'wb') as private_key_file:
            private_key_file.write(crypto.dump_privatekey(FILETYPE_PEM, self.key))

    def _generate_cert(self, service_id):
        csr = self._generate_csr(service_id)

        cert = self.cert_client.send_csr(service_id=service_id, csr=csr, cert_type=self.cert_type)

        cert_path = self._service_cert_path(service_id)
        logger.info(f'Saving service certificate: {cert_path}')
        with open(cert_path, 'w') as cert_file:
            cert_file.write(cert)

        return crypto.load_certificate(FILETYPE_PEM, cert)

    def _generate_csr(self, service_id):
        logger.info('Generating csr')

        csr = X509Req()
        csr.get_subject().C = 'PL'
        csr.get_subject().ST = 'Poland'
        csr.get_subject().O = 'REXS.IO'
        csr.get_subject().CN = f'{self.cert_type}-{service_id}'

        csr.set_pubkey(self.key)
        csr.sign(self.key, 'sha256')

        return crypto.dump_certificate_request(FILETYPE_PEM, csr)

    def _generate_p12_cert(self, cert, service_id):
        p12_cert_path = self._p12_cert_path(service_id)
        logger.info(f'Generating p12 certificate: {p12_cert_path}')

        pkcs12_cert = PKCS12()
        pkcs12_cert.set_certificate(cert)
        pkcs12_cert.set_ca_certificates(self.ca_certs)
        pkcs12_cert.set_privatekey(self.key)

        with open(p12_cert_path, 'wb') as pkcs12_cert_file:
            pkcs12_cert_file.write(pkcs12_cert.export(KNOWN_PASSWORD))

    def _generate_truststore(self, service_id):
        truststore_path = self._truststore_path(service_id)
        logger.info(f'Generating truststore: {truststore_path}')

        truststore = PKCS12()
        truststore.set_certificate(self.ca_certs[0])
        truststore.set_friendlyname(b'rexsio')

        with open(truststore_path, 'wb') as truststore_file:
            truststore_file.write(truststore.export(KNOWN_PASSWORD))

    def _private_key_path(self, service_id):
        return self._service_config_file_path(service_id, 'service-private.key')

    def _service_cert_path(self, service_id):
        return self._service_config_file_path(service_id, 'service.crt')

    def _ca_certs_chain_path(self, service_id):
        return self._service_config_file_path(service_id, 'ca-certificates.chain')

    def _truststore_path(self, service_id):
        return self._service_config_file_path(service_id, 'truststore.p12')

    def _p12_cert_path(self, service_id):
        return self._service_config_file_path(service_id, 'service.p12')

    def _service_config_file_path(self, service_id, file_name):
        return f'{self._service_config_path(service_id)}/{file_name}'

    def _service_config_path(self, service_id):
        return f'{self.service_configs_path}/{service_id}/config/{self.cert_type.lower()}'



import json
from typing import Dict, Any

from agent.rabbitmq import logger
from agent.rabbitmq.client import ReconnectingRabbitMqClient


def publish_message(message: Dict[str, Any]) -> None:
    rabbitmq_message = json.dumps(message).encode("utf-8")
    rabbitmq_client = ReconnectingRabbitMqClient.get_instance().client
    rabbitmq_client.publish_message(
        message=rabbitmq_message,
        message_type=message["messageType"],
        topic=message["nodeServiceId"],
    )
    logger.info(f"Message sent to RabbitMQ: {rabbitmq_message}")



from agent.commands.metrics import read_metrics
from agent.scheduler import logger
from agent.scheduler.scheduled_task import ScheduledTask
from agent.websocket.client import AgentWebSocketClient


class MetricsTask(ScheduledTask):

    def execute(self):
        metrics = read_metrics()
        AgentWebSocketClient.get_instance().safe_message_send(metrics)
        logger.info(f'Node metrics sent: {metrics}')



import shutil

from agent.commands.utils import get_field_from_body, get_base_service_path
from agent.commands.docker_commands import docker_compose_down
from agent.constants import DOCKER_COMPOSE_FILE
from agent.scheduler.scheduled_tasks_manager import ScheduledTaskManager


def delete_service(body):
    service_id = get_field_from_body('nodeServiceId', body)
    service_path = get_base_service_path(service_id)
    docker_compose_file = f'{service_path}/{DOCKER_COMPOSE_FILE}'

    tasks_manager = ScheduledTaskManager.get_instance()
    tasks_manager.stop_executing()
    docker_compose_down(docker_compose_file)
    shutil.rmtree(service_path, ignore_errors=True)
    tasks_manager.start_executing()



import functools
import threading
import time

import pika

from agent.config.properties import RABBITMQ_URL
from agent.constants import NOTIFY_CONTROL_CENTER, EXCHANGE, EXCHANGE_TYPE
from agent.rabbitmq import logger


class RabbitMqClient(object):
    def __init__(self):
        self.should_reconnect = False
        self.was_consuming = False
        self.should_stop = False

        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._url = RABBITMQ_URL
        self._consuming = False
        self._prefetch_count = 1

    def connect(self):
        logger.info(f"Connecting to {self._url}")
        return pika.SelectConnection(
            parameters=pika.URLParameters(self._url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
        )

    def close_connection(self):
        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            logger.info("Connection is closing or already closed")
        else:
            logger.info("Closing connection")
            self._connection.close()

    def on_connection_open(self, _unused_connection):
        logger.info("Connection opened")
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        logger.error(f"Connection open failed: {err}")
        self.reconnect()

    def on_connection_closed(self, _unused_connection, reason):
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            logger.warning(f"Connection closed, reconnect necessary: {reason}")
            self.reconnect()

    def reconnect(self):
        self.should_reconnect = True
        self.stop()

    def open_channel(self):
        logger.info("Creating a new channel")
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        logger.info("Channel opened")
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(EXCHANGE)

    def add_on_channel_close_callback(self):
        logger.info("Adding channel close callback")
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        logger.warning(f"Channel {channel} was closed: {reason}")
        self.close_connection()

    def setup_exchange(self, exchange_name):
        logger.info(f"Declaring exchange: {exchange_name}")
        cb = functools.partial(self.on_exchange_declareok, userdata=exchange_name)
        self._channel.exchange_declare(
            exchange=exchange_name, exchange_type=EXCHANGE_TYPE, callback=cb
        )

    def on_exchange_declareok(self, _unused_frame, userdata):
        logger.info(f"Exchange declared: {userdata}")
        self.setup_queue(NOTIFY_CONTROL_CENTER)

    def setup_queue(self, queue_name):
        logger.info(f"Declaring queue {queue_name}")
        cb = functools.partial(self.on_queue_declareok, userdata=queue_name)
        self._channel.queue_declare(queue=queue_name, callback=cb)

    def on_queue_declareok(self, _unused_frame, userdata):
        queue_name = userdata
        logger.info(f"Binding {EXCHANGE} to {queue_name} with {NOTIFY_CONTROL_CENTER}")
        cb = functools.partial(self.on_bindok, userdata=queue_name)
        self._channel.queue_bind(
            queue_name, EXCHANGE, routing_key=NOTIFY_CONTROL_CENTER, callback=cb
        )

    def on_bindok(self, _unused_frame, userdata):
        logger.info(f"Queue bound: {userdata}")
        self.set_qos()

    def set_qos(self):
        self._channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
        )

    def on_basic_qos_ok(self, _unused_frame):
        logger.info(f"QOS set to: {self._prefetch_count}")
        self.start_consuming()

    def start_consuming(self):
        logger.info("Issuing consumer related RPC commands")
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(
            NOTIFY_CONTROL_CENTER, self.on_message
        )
        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_callback(self):
        logger.info("Adding consumer cancellation callback")
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        logger.info(f"Consumer was cancelled remotely, shutting down: {method_frame}")
        if self._channel:
            self._channel.close()

    def on_message(self, _unused_channel, basic_deliver, _properties, body):
        logger.info(f"Received message # {basic_deliver.delivery_tag}: {body}")
        self.notify_control_center(body)
        self.acknowledge_message(basic_deliver.delivery_tag)

    @staticmethod
    def notify_control_center(body: bytes) -> None:
        from agent.websocket.client import AgentWebSocketClient

        AgentWebSocketClient.get_instance().safe_message_send(body)

    def acknowledge_message(self, delivery_tag):
        logger.info(f"Acknowledging message {delivery_tag}")
        self._channel.basic_ack(delivery_tag)

    def stop_consuming(self):
        if self._channel:
            logger.info("Sending a Basic.Cancel RPC command to RabbitMQ")
            cb = functools.partial(self.on_cancelok, userdata=self._consumer_tag)
            self._channel.basic_cancel(self._consumer_tag, cb)

    def on_cancelok(self, _unused_frame, userdata):
        self._consuming = False
        logger.info(
            f"RabbitMQ acknowledged the cancellation of the consumer: {userdata}"
        )
        self.close_channel()

    def close_channel(self):
        logger.info("Closing the channel")
        self._channel.close()

    def run(self):
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        if not self._closing:
            self._closing = True
            logger.info("Stopping")
            if self._consuming:
                self.stop_consuming()
                self._connection.ioloop.start()
            else:
                self._connection.ioloop.stop()
            logger.info("Stopped")

    def publish_message(self, message: bytes, message_type: str, topic: str) -> None:
        self._channel.queue_bind(exchange=EXCHANGE, queue=topic, routing_key=topic)
        self._channel.basic_publish(
            exchange=EXCHANGE,
            routing_key=topic,
            body=message,
            properties=pika.BasicProperties(
                content_type="application/json", type=message_type
            ),
        )


class ReconnectingRabbitMqClient(object):
    _instance = None

    @staticmethod
    def get_instance():
        if ReconnectingRabbitMqClient._instance is None:
            ReconnectingRabbitMqClient._instance = ReconnectingRabbitMqClient()
        return ReconnectingRabbitMqClient._instance

    def __init__(self):
        self._reconnect_delay = 0
        self._stop_consuming_event = threading.Event()
        self.client = RabbitMqClient()
        self.thread = None

    def start_consuming_messages(self):
        logger.info("Starting consuming messages on RabbitMQ")
        self.thread = threading.Thread(
            target=self._run, args=(self._stop_consuming_event,)
        )
        self.thread.start()

    def stop_consuming_messages(self):
        logger.info("Stopping consuming messages on RabbitMQ")
        self._stop_consuming_event.set()
        self.client.stop_consuming()
        self.thread.join()

    def _run(self, stop_event: threading.Event) -> None:
        while not stop_event.isSet():
            self.client.run()
            self._maybe_reconnect()

    def _maybe_reconnect(self):
        if self.client.should_reconnect:
            self.client.stop()
            reconnect_delay = self._get_reconnect_delay()
            logger.info(f"Reconnecting after {reconnect_delay} seconds")
            time.sleep(reconnect_delay)
            self.client = RabbitMqClient()

    def _get_reconnect_delay(self):
        if self.client.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay



from agent.commands.service_restart import restart_services
from agent.commands.service_status import check_service_statuses
from agent.scheduler.scheduled_task import ScheduledTask
from agent.services.status_sender import send_statuses


class StatusTask(ScheduledTask):

    def execute(self):
        statuses_to_send, services_to_restart = check_service_statuses()
        send_statuses(statuses_to_send)
        restart_services(services_to_restart)



import re
import shlex
import subprocess

from agent.constants import FILTER_SMART_CARD_COMMANDS, LIST_USB_DEVICES
from agent.exceptions.agent_error import AgentError

address_re = re.compile("Bus\s+(?P<bus>\d+)\s+Device\s+(?P<device>\d+).+ID\s.+$", re.I)


def find_smart_card_reader_address():
    raw_reader_address = _get_reader_usb_details()
    return _parse_reader_address(raw_reader_address)


def _get_reader_usb_details():
    input_process = _open_subprocess_pipe(command=LIST_USB_DEVICES)
    for command in FILTER_SMART_CARD_COMMANDS:
        current_process = _open_subprocess_pipe(command=command,
                                                stdin=input_process.stdout)
        input_process.stdout.close()
        input_process = current_process
    filtered_devices, error = input_process.communicate()
    if error is not None:
        raise AgentError(expression=filtered_devices, message=error)
    return filtered_devices


def _open_subprocess_pipe(command, **kwargs):
    return subprocess.Popen(args=shlex.split(command),
                            stdout=subprocess.PIPE,
                            universal_newlines=True,
                            **kwargs)


def _parse_reader_address(raw_data):
    data_lines = raw_data.split('\n')
    address_line = _get_line_with_the_usb_address(data_lines)
    bus_and_device = address_re.match(address_line).groupdict()
    return _build_usb_address(bus_and_device)


def _get_line_with_the_usb_address(data_lines):
    return next(data_line for data_line in data_lines if 'Bus' and 'Device' in data_line)


def _build_usb_address(bus_and_device):
    bus = bus_and_device.pop('bus')
    device = bus_and_device.pop('device')
    return f"/dev/bus/usb/{bus}/{device}"



import subprocess

from agent.commands import logger
from agent.commands.utils import get_docker_compose_path, get_field_from_body
from agent.commands.docker_commands import docker_compose_restart
from agent.constants import SERVICE_RESTARTING, SERVICE_DOWN, SERVICE_UP
from agent.services.status_sender import send_status


def restart_service(body):
    service_id = get_field_from_body('nodeServiceId', body)
    logger.info(f'Restarting {service_id} service')
    _restart_service(service_id)


def restart_services(service_id_list):
    for service_id in service_id_list:
        logger.info(f'Service {service_id} is DOWN, restarting')
        _restart_service(service_id)


def _restart_service(service_id):
    send_status(service_id, SERVICE_RESTARTING)
    docker_compose_path = get_docker_compose_path(service_id)
    try:
        docker_compose_restart(docker_compose_path)
        send_status(service_id, SERVICE_UP)
    except subprocess.CalledProcessError:
        send_status(service_id, SERVICE_DOWN)



import docker

from agent.commands.utils import prepare_message_to_send
from agent.constants import SERVICE_DOWN, SERVICE_UP, SERVICES_STATUS
from agent.services.utils import get_services_id_list


def check_service_statuses():
    service_id_list = get_services_id_list()
    client = docker.from_env()
    service_status_list = _get_services_status_list(service_id_list, client.containers)
    services_to_restart = _get_services_id_to_restart(service_status_list)
    service_status_body = {"statuses":service_status_list}
    service_status_message = prepare_message_to_send(message_type=SERVICES_STATUS, body=service_status_body)
    return service_status_message, services_to_restart


def _get_services_status_list(id_list, containers):
    statuses_list = [_get_service_status(containers, service_id) for service_id in id_list]
    services_statuses = [dict(serviceId=service_id, status=status) for service_id, status in
                         list(zip(id_list, statuses_list))]
    return _remove_services_with_none_status(services_statuses)


def _remove_services_with_none_status(services_statuses):
    services_without_none_status = []
    for service in services_statuses:
        if service['status'] is not None:
            services_without_none_status.append(service)
    return services_without_none_status


def _get_service_status(containers, service_id):
    service_filter = service_id.lower()
    service_containers = containers.list(all=True, filters=dict(name=service_filter))
    if not service_containers:
        service_containers = containers.list(all=True, filters=dict(name=service_filter.replace('-', '')))
    if not service_containers:
        return None
    service_containers_status = [container.status for container in service_containers]
    return SERVICE_UP if _are_all_containers_running(service_containers_status) else SERVICE_DOWN


def _are_all_containers_running(service_containers_status):
    return all(container_status == 'running' for container_status in service_containers_status)


def _get_services_id_to_restart(service_status_list):
    return [service['serviceId'] for service in service_status_list if service['status'] == SERVICE_DOWN]



from collections import namedtuple

from agent.config.properties import ACCESS_TOKEN_PATH, REXSIO_WS_HOST

command = namedtuple("command", ["type", "body"])


def resolve_ws_endpoint():
    access_token = get_access_token()
    return f"wss://{REXSIO_WS_HOST}/ws?access_token={access_token}"


def get_access_token():
    with open(ACCESS_TOKEN_PATH, "r") as f:
        token = f.read()
    return token.strip()



import socket

import psutil

from agent.commands import logger
from agent.commands.utils import prepare_message_to_send
from agent.constants import NODE_METRICS
from agent.exceptions.agent_error import AgentError


def read_metrics():
    ip_address = _get_network_interface_controller_ip_address()
    metrics_body = dict(ip=ip_address)
    return prepare_message_to_send(message_type=NODE_METRICS, body=metrics_body)


def _get_network_interface_controller_ip_address():
    network_interfaces = psutil.net_if_addrs()
    try:
        ethernet_interface = _find_ethernet_interface(network_interfaces)
        snic_addresses = network_interfaces.get(ethernet_interface)
        snic = _find_af_inet_snic(snic_addresses)
        return snic.address
    except AgentError as err:
        logger.error(f'IP address not found. {err}')


def _find_ethernet_interface(network_interfaces):
    interfaces = network_interfaces.keys()
    try:
        return next(interface for interface in interfaces if interface.startswith('e'))
    except StopIteration as err:
        raise AgentError(expression=interfaces, message=err)


def _find_af_inet_snic(snic_addresses):
    try:
        return next(snic for snic in snic_addresses if snic.family == socket.AF_INET)
    except StopIteration as err:
        raise AgentError(expression=snic_addresses, message=err)



from twisted.internet.task import LoopingCall

from agent.scheduler import logger
from agent.scheduler.metrics_task import MetricsTask
from agent.scheduler.status_task import StatusTask

TASKS = [
    StatusTask(),
    MetricsTask()
]


class ScheduledTaskManager:

    _instance = None

    @staticmethod
    def get_instance():
        if ScheduledTaskManager._instance is None:
            ScheduledTaskManager._instance = ScheduledTaskManager()
        return ScheduledTaskManager._instance

    def __init__(self):
        self.scheduled_tasks = list(map(lambda t: {'task': t, 'looping_call': LoopingCall(t.execute)}, TASKS))

    def start_executing(self):
        for scheduled_task in self.scheduled_tasks:
            if not scheduled_task["looping_call"].running:
                logger.info(f'Starting executing {scheduled_task["task"].__class__.__name__}...')
                scheduled_task["looping_call"].start(scheduled_task["task"].get_interval_seconds())

    def stop_executing(self):
        for scheduled_task in self.scheduled_tasks:
            if scheduled_task["looping_call"].running:
                logger.info(f'Stopping executing {scheduled_task["task"].__class__.__name__}...')
                scheduled_task["looping_call"].stop()



import abc


class ScheduledTask(abc.ABC):
    @abc.abstractmethod
    def execute(self):
        """
        This method will be executed with intervals defined by #get_interval_seconds method
        """
        pass

    def get_interval_seconds(self):
        """
        Defines interval

        By default 20
        """
        return 20



import os
from collections import namedtuple
from typing import Dict
from urllib.error import HTTPError

import requests

from agent.certificates.certificate_client import CertificateClient
from agent.certificates.certificate_config_generator import CertificateConfigGenerator
from agent.commands import logger
from agent.commands.docker_commands import docker_compose_pull, docker_compose_up
from agent.commands.smart_card_reader import find_smart_card_reader_address
from agent.commands.utils import save_config, get_field_from_body, get_docker_compose_path
from agent.config.properties import BASE_SERVICES_PATH, REXSIO_HOST, RABBITMQ_PASS
from agent.exceptions.agent_error import AgentError
from agent.websocket.utils import get_access_token


def provision_service(body):
    service = _get_service_data_from_json(body)
    _add_rabbitmq_pass_to_config(service.config)
    save_config(service_id=service.id, config=service.config)
    _generate_service_certificate(service.id)
    docker_compose_file_path = _download_docker_compose(compose_url=service.compose_url, service_id=service.id)
    _set_smart_card_reader_address_if_needed(service_type=service.type, config=service.config)
    docker_compose_pull(docker_compose_file_path)
    docker_compose_up(docker_compose_file_path)
    logger.info(f'Service {service.type}/{service.id} is running')


def _add_rabbitmq_pass_to_config(config: Dict[str, str]) -> None:
    config["rabbitmqPass"] = RABBITMQ_PASS


def _get_service_data_from_json(body):
    service = namedtuple('service', ['type', 'id', 'config', 'compose_url', 'version'])
    service.type = get_field_from_body('serviceType', body)
    service.id = get_field_from_body('nodeServiceId', body)
    service.config = get_field_from_body('config', body)
    service.compose_url = get_field_from_body('dockerComposeUrl', body)
    service.version = get_field_from_body('version', body)
    return service


def _generate_service_certificate(service_id):
    token = get_access_token()
    rexsio_path = f'https://{REXSIO_HOST}'
    client = CertificateClient(base_path=rexsio_path, access_token=token)
    generator = CertificateConfigGenerator(service_configs_path=BASE_SERVICES_PATH,
                                           cert_client=client)
    generator.generate_cert_config(service_id)


def _download_docker_compose(compose_url, service_id):
    file_path = get_docker_compose_path(service_id)
    try:
        _get_docker_compose_file(url=compose_url, filename=file_path)
    except HTTPError as error:
        logger.error(f'Docker compose file download failed: {file_path}, url: {compose_url}')
        raise AgentError(expression=compose_url, message=error)
    logger.info(f'Docker compose file downloaded: {file_path}')
    return file_path


def _get_docker_compose_file(url, filename):
    r = requests.get(url)
    r.raise_for_status()
    with open(filename, 'wb') as file:
        file.write(r.content)


def _set_smart_card_reader_address_if_needed(service_type, config):
    if _is_card_reader_needed(service_type, config):
        _set_smart_card_reader_address()


def _set_smart_card_reader_address():
    address = find_smart_card_reader_address()
    os.environ['REXSIO_SMART_CARD_READER_ADDRESS'] = address
    logger.info(f'Smart card reader address added: {address}')


def _is_card_reader_needed(service_type, config):
    return _is_service_notary(service_type) and _is_not_using_private_key(config)


def _is_service_notary(service_type):
    return 'notary' in service_type.lower()


def _is_not_using_private_key(config):
    is_private_key = get_field_from_body('usePrivateKey', config)
    return not is_private_key



# TODO: https://gitlab.com/bettersolutions/issues-dac/rexs.io/-/issues/279
try:
    from importlib.resources import path
except ImportError:
    from importlib_resources import path

from agent.commands.docker_commands import docker_compose_up, docker_compose_pull
from agent.constants import DOCKER_COMPOSE_FILE


def setup_rabbitmq():
    with path("agent.rabbitmq.rabbit_docker_compose", f"{DOCKER_COMPOSE_FILE}") as docker_compose_path:
        docker_compose_pull(path=docker_compose_path)
        docker_compose_up(path=docker_compose_path)



import logging

from twisted.python import log

LOGGER_ROOT = 'agent'


def setup_logger():
    observer = log.PythonLoggingObserver(LOGGER_ROOT)
    observer.start()

    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(name)s - %(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(LOGGER_ROOT)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    return logger


def get_logger(name):
    return logging.getLogger('.'.join([LOGGER_ROOT, name]))



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rexsio-agent",
    version=os.getenv('RELEASE_VERSION'),
    entry_points={
        'console_scripts': [
            'agent = agent.__main__:main'
        ]
    },
    author="rexs.io sp. z o.o.",
    author_email="hello@rexs.io",
    description="Rexsio Agent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/rexsio/agent",
    packages=setuptools.find_packages(),
    include_package_data=True,
    test_suite="tests",
    tests_require=[],
    install_requires=['autobahn==20.2.1',
                      'Twisted==19.10.0',
                      'requests==2.22.0',
                      'psutil==5.6.7',
                      'docker==4.2.0',
                      'pyOpenSSL==19.1.0',
                      'service_identity==18.1.0',
                      'python-dotenv==0.12.0',
                      'pika==1.1.0',
                      'importlib_resources==3.0.0'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



from agent.commands import logger
from agent.commands.service_delete import delete_service
from agent.commands.service_provision import provision_service
from agent.commands.service_restart import restart_service
from agent.commands.uninstall import uninstall
from agent.constants import (
    PROVISION_SERVICE,
    DELETE_SERVICE,
    DELETE_NODE,
    RESTART_SERVICE,
)
from agent.exceptions.agent_error import AgentError


def dispatch_command(command, body):
    _validate_command(command)

    if not body:
        return commands_dispatcher[command]()

    _validate_body(body)
    return commands_dispatcher[command](body)


def _validate_command(command):
    if command in commands_dispatcher.keys():
        logger.info(f"Recognized command: {command}!")
    else:
        raise AgentError(command, f"Invalid command!")


def _validate_body(body):
    if isinstance(body, dict):
        logger.info(f"{body} is a dict!")
    else:
        raise AgentError(body, f"Body is not a dict!")


commands_dispatcher = {
    PROVISION_SERVICE: provision_service,
    DELETE_SERVICE: delete_service,
    DELETE_NODE: uninstall,
    RESTART_SERVICE: restart_service,
}



import os

from dotenv import load_dotenv, find_dotenv

load_dotenv(find_dotenv())

INSTALL_DIR = os.getenv("INSTALL_DIR", "/opt/rexsio")
ROOT_DIR = f"{INSTALL_DIR}/agent"
BASE_SERVICES_PATH = f"{INSTALL_DIR}/services"
ACCESS_TOKEN_PATH = f"{ROOT_DIR}/access.token"
REXSIO_HOST = os.getenv("REXSIO_HOST", "api-dev.rexs.io")
REXSIO_WS_HOST = os.getenv("REXSIO_WS_HOST", "ws-service.rexsio.dev.dac.systems")
REXSIO_USER = os.getenv("REXSIO_USER", "rexsio")
RABBITMQ_PASS = os.getenv("RABBITMQ_PASS", "rexsio")
RABBITMQ_BASE_URL = os.getenv("RABBITMQ_BASE_URL", "localhost")
RABBITMQ_URL = f"amqp://rexsio:{RABBITMQ_PASS}@{RABBITMQ_BASE_URL}:5672"



import shlex
import subprocess

from agent.config.properties import INSTALL_DIR, REXSIO_USER
from agent.constants import UNINSTALL_SCRIPT_FILE
from agent.exceptions.agent_error import AgentError
from agent.scheduler.scheduled_tasks_manager import ScheduledTaskManager


def uninstall():
    ScheduledTaskManager.get_instance().stop_executing()
    uninstall_script_path = f'{INSTALL_DIR}/{UNINSTALL_SCRIPT_FILE}'
    try:
        subprocess.call(shlex.split(f'sudo {uninstall_script_path} {INSTALL_DIR} {REXSIO_USER}'))
    except subprocess.CalledProcessError as err:
        raise AgentError(expression=uninstall_script_path, message=err.output)



from agent.commands.utils import prepare_message_to_send
from agent.constants import SERVICES_STATUS
from agent.scheduler import logger
from agent.websocket.client import AgentWebSocketClient


def send_statuses(statuses):
    AgentWebSocketClient.get_instance().safe_message_send(statuses)
    logger.info(f'Service statuses sent: {statuses}')


def send_status(service_id, status):
    service_status_list = {"statuses":[{"serviceId": service_id, "status": status}]}
    service_status_message = prepare_message_to_send(message_type=SERVICES_STATUS, body=service_status_list)
    send_statuses(service_status_message)



PROVISION_SERVICE = 'PROVISION_SERVICE'
DELETE_SERVICE = 'DELETE_SERVICE'
NODE_METRICS = 'NODE_METRICS'
SERVICES_STATUS = 'SERVICES_STATUS'
DELETE_NODE = 'DELETE_NODE'
RESTART_SERVICE = 'RESTART_SERVICE'

SERVICE_UP = 'UP'
SERVICE_DOWN = 'DOWN'
SERVICE_RESTARTING = 'RESTARTING'

LIST_USB_DEVICES = 'lsusb -v'
FILTER_SMART_CARD_COMMANDS = ['grep -E "\<(ID|Chip/SmartCard)"', 'grep "Chip/SmartCard" -B 1']

UNINSTALL_SCRIPT_FILE = 'uninstall.sh'

DOCKER_COMPOSE_FILE = 'docker-compose.yaml'

EXCHANGE = "agent"
EXCHANGE_TYPE = "topic"
NOTIFY_CONTROL_CENTER = 'NOTIFY_CONTROL_CENTER'
